"use strict";
/**
 * @module Security
 */
Object.defineProperty(exports, "__esModule", { value: true });
var util_1 = require("features/security/util");
var require_non_null_1 = require("util/require-non-null");
/**
 * Implementation of the {@link SecurityScriptBuilder} interface
 */
var SecurityScriptBuilderImpl = /** @class */ (function () {
    /**
     * Create a new script builder
     *
     * @param commands  the security script built so far
     */
    function SecurityScriptBuilderImpl(commands) {
        if (commands === void 0) { commands = ''; }
        this.commands = commands;
    }
    /**
     * Utility function that creates a new SecurityScriptBuilderImpl with an
     * additional command appended
     *
     * @param command  the command to append
     * @return         a new security script builder
     */
    SecurityScriptBuilderImpl.prototype.withCommand = function (command) {
        if (this.commands !== '') {
            return new SecurityScriptBuilderImpl(this.commands + '\n' + command);
        }
        return new SecurityScriptBuilderImpl(command);
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.build = function () {
        return this.commands;
    };
    /**
     * Convert the security script builder to a string
     *
     * @return  a string representation on this object
     */
    SecurityScriptBuilderImpl.prototype.toString = function () {
        return "ConfigurationScriptBuilder [" + this.commands + "]";
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setRolesForAnonymousSessions = function (roles) {
        return this.withCommand('set roles for anonymous sessions ' +
            util_1.rolesToString(require_non_null_1.requireNonNull(roles, 'roles')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setRolesForNamedSessions = function (roles) {
        return this.withCommand('set roles for named sessions ' +
            util_1.rolesToString(require_non_null_1.requireNonNull(roles, 'roles')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setGlobalPermissions = function (role, permissions) {
        return this.withCommand('set ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(role, 'role')) +
            ' permissions ' +
            util_1.permissionsToString(require_non_null_1.requireNonNull(permissions, 'permissions')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setDefaultTopicPermissions = function (role, permissions) {
        return this.withCommand('set ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(role, 'role')) +
            ' default topic permissions ' +
            util_1.permissionsToString(require_non_null_1.requireNonNull(permissions, 'permissions')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.removeTopicPermissions = function (role, path) {
        return this.withCommand('remove ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(role, 'role')) +
            ' permissions for topic ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(path, 'path')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setTopicPermissions = function (role, path, permissions) {
        return this.withCommand('set ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(role, 'role')) +
            ' topic ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(path, 'path')) +
            ' permissions ' +
            util_1.permissionsToString(require_non_null_1.requireNonNull(permissions, 'permissions')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setRoleIncludes = function (role, included) {
        return this.withCommand('set ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(role, 'role')) +
            ' includes ' +
            util_1.rolesToString(require_non_null_1.requireNonNull(included, 'included roles')));
    };
    /**
     * @inheritdoc
     */
    SecurityScriptBuilderImpl.prototype.setRoleLockedByPrincipal = function (role, lockingPrincipal) {
        return this.withCommand('set role ' + util_1.quoteAndEscape(require_non_null_1.requireNonNull(role, 'role is null')) +
            ' locked by principal ' + util_1.quoteAndEscape(require_non_null_1.requireNonNull(lockingPrincipal, 'lockingPrincipal is null')));
    };
    return SecurityScriptBuilderImpl;
}());
exports.SecurityScriptBuilderImpl = SecurityScriptBuilderImpl;
