"use strict";
/**
 * @module Security
 */
Object.defineProperty(exports, "__esModule", { value: true });
var util_1 = require("features/security/util");
var require_non_null_1 = require("util/require-non-null");
/**
 * Implementation of the {@link SystemAuthenticationScriptBuilder} interface
 */
var SystemAuthenticationScriptBuilderImpl = /** @class */ (function () {
    /**
     * Create a new script builder
     *
     * @param commands  the security script built so far
     */
    function SystemAuthenticationScriptBuilderImpl(commands) {
        if (commands === void 0) { commands = ''; }
        this.commands = commands;
    }
    /**
     * Utility function that creates a new SystemAuthenticationScriptBuilderImpl with an
     * additional command appended
     *
     * @param command  the command to append
     * @return         a new system authentication script builder
     */
    SystemAuthenticationScriptBuilderImpl.prototype.withCommand = function (command) {
        return new SystemAuthenticationScriptBuilderImpl(this.commands !== ''
            ? this.commands + '\n' + command
            : command);
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.build = function () {
        return this.commands;
    };
    /**
     * Convert the system authentication script builder to a string
     *
     * @return  a string representation on this object
     */
    SystemAuthenticationScriptBuilderImpl.prototype.toString = function () {
        return "SystemAuthenticationScriptBuilder [" + this.commands + "]";
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.assignRoles = function (principal, roles) {
        return this.withCommand('assign roles ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(principal, 'principal')) +
            ' ' +
            util_1.rolesToString(require_non_null_1.requireNonNull(roles, 'roles')));
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.addPrincipal = function (principal, password, roles, lockingPrincipal) {
        if (roles === void 0) { roles = []; }
        return this.withCommand('add principal ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(principal, 'principal')) +
            ' ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(password, 'password')) +
            (roles.length ? ' ' + util_1.rolesToString(roles) : '') +
            (lockingPrincipal ? ' locked by ' + util_1.quoteAndEscape(lockingPrincipal) : ''));
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.setPassword = function (principal, password) {
        return this.withCommand('set password ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(principal, 'principal')) +
            ' ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(password, 'password')));
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.verifyPassword = function (principal, password) {
        return this.withCommand('verify password ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(principal, 'principal')) +
            ' ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(password, 'password')));
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.removePrincipal = function (principal) {
        return this.withCommand('remove principal' +
            ' ' +
            util_1.quoteAndEscape(require_non_null_1.requireNonNull(principal, 'principal')));
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.allowAnonymousConnections = function (roles) {
        roles = roles || [];
        return this.withCommand('allow anonymous connections' +
            ' ' +
            util_1.rolesToString(roles));
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.denyAnonymousConnections = function () {
        return this.withCommand('deny anonymous connections');
    };
    /**
     * @inheritdoc
     */
    SystemAuthenticationScriptBuilderImpl.prototype.abstainAnonymousConnections = function () {
        return this.withCommand('abstain anonymous connections');
    };
    return SystemAuthenticationScriptBuilderImpl;
}());
exports.SystemAuthenticationScriptBuilderImpl = SystemAuthenticationScriptBuilderImpl;
