"use strict";
/**
 * @module Topics
 */
Object.defineProperty(exports, "__esModule", { value: true });
var Codec = require("io/codec");
var BEES = require("serialisers/byte-encoded-enum-serialiser");
var RecordTopicDetailsSerialiser = require("topics/details/record-topic-details-serialiser");
var RoutingTopicDetailsSerialiser = require("topics/details/routing-topic-details-serialiser");
var SlaveTopicDetailsSerialiser = require("topics/details/slave-topic-details-serialiser");
var TopicPropertiesSerialiser = require("topics/details/topic-properties-serialiser");
var topic_type_1 = require("../../../topics/topic-type");
/**
 * Get a special serialiser for {@link TopicDetailsAttributes} for Record,
 * Routing and Slave topics
 *
 * @param  type  the topic type
 * @return       the special serialiser, or `null` if no special serialiser is
 *               needed
 */
function getSubSerialiser(type) {
    switch (type) {
        case topic_type_1.TopicTypeEnum.RECORD:
            return RecordTopicDetailsSerialiser;
        case topic_type_1.TopicTypeEnum.ROUTING:
            return RoutingTopicDetailsSerialiser;
        case topic_type_1.TopicTypeEnum.SLAVE:
            return SlaveTopicDetailsSerialiser;
    }
    return null;
}
/**
 * Read {@link TopicDetails} from the stream
 *
 * @param bis the input stream
 * @return    the {@link TopicDetails} that were read
 */
function read(bis) {
    var details = {};
    if (Codec.readBoolean(bis)) {
        details.type = BEES.read(bis, topic_type_1.TopicType);
        if (Codec.readBoolean(bis) === false) {
            return details;
        }
        details.schema = {};
        if (Codec.readBoolean(bis) === false) {
            return details;
        }
        // auto subscribe (ignored)
        Codec.readBoolean(bis);
        var tidyOnUnsubscribe = Codec.readBoolean(bis);
        var reference = Codec.readString(bis);
        var properties = TopicPropertiesSerialiser.read(bis);
        var serialiser = getSubSerialiser(details.type);
        var attributes = {};
        if (serialiser) {
            attributes = serialiser.read(bis);
        }
        attributes.tidyOnUnsubscribe = tidyOnUnsubscribe;
        attributes.properties = properties;
        attributes.reference = reference;
        details.attributes = attributes;
    }
    return details;
}
exports.read = read;
/**
 * Write a {@link TopicSpecification} to the stream
 *
 * @param bos   the output stream
 * @param value the {@link TopicSpecification} to be written
 */
function write(bos, value) {
    if (value) {
        Codec.writeBoolean(bos, true);
        BEES.write(bos, value.type);
        // schema
        Codec.writeBoolean(bos, true);
        // have attributes, write flag
        Codec.writeBoolean(bos, true);
        // auto subscribe (redundant)
        Codec.writeBoolean(bos, true);
        if (value.attributes === undefined) {
            throw new Error('Topic Details Attributes is undefined');
        }
        // general attributes
        Codec.writeBoolean(bos, value.attributes.tidyOnUnsubscribe);
        Codec.writeString(bos, value.attributes.reference);
        TopicPropertiesSerialiser.write(bos, value.attributes.properties);
        var serialiser = getSubSerialiser(value.type);
        if (serialiser) {
            serialiser.write(bos, value.attributes);
        }
    }
    else {
        // no details
        Codec.writeBoolean(bos, false);
    }
}
exports.write = write;
